from collections import OrderedDict

from django.contrib.contenttypes.fields import GenericForeignKey


def get_calculation_annotation(calculation_field, calculation_method):
    """
    Returns the default django annotation
    @param calculation_field: the field to calculate ex 'value'
    @param calculation_method: the aggregation method ex: Sum
    @return: the annotation ex value__sum
    """

    return "__".join([calculation_field.lower(), calculation_method.name.lower()])


def get_foreign_keys(model):
    """
    Scans a model and return an Ordered Dictionary with the foreign keys found
    :param model: the model to scan
    :return: Ordered Dict
    """
    from django.db import models

    fields = model._meta.get_fields()
    fkeys = OrderedDict()
    for f in fields:
        if (
            f.is_relation
            and type(f) is not models.OneToOneRel
            and type(f) is not models.ManyToOneRel
            and type(f) is not models.ManyToManyRel
            and type(f) is not GenericForeignKey
        ):
            fkeys[f.attname] = f
    return fkeys


def get_field_from_query_text(path, model):
    """
    return the field of a query text
    `modelA__modelB__foo_field` would return foo_field on modelsB
    :param path:
    :param model:
    :return:
    """
    relations = path.split("__")
    _rel = model
    field = None
    for i, m in enumerate(relations):
        field = _rel._meta.get_field(m)
        if i == len(relations) - 1:
            return field
        _rel = field.related_model
    return field



from django import template
from django.template.loader import get_template
from django.forms import Media
from django.urls import reverse, resolve
from django.utils.safestring import mark_safe

from ..app_settings import SLICK_REPORTING_JQUERY_URL, SLICK_REPORTING_SETTINGS, get_media

register = template.Library()


@register.simple_tag
def get_widget_from_url(url_name=None, url=None, **kwargs):
    _url = ""
    if not (url_name or url):
        raise ValueError("url_name or url must be provided")
    if url_name:
        url = reverse(url_name)
    view = resolve(url)
    kwargs["report"] = view.func.view_class
    kwargs["report_url"] = url
    return get_widget(**kwargs)


@register.simple_tag
def get_widget(report, template_name="", url_name="", report_url=None, **kwargs):
    kwargs["report"] = report
    if not report:
        raise ValueError("report argument is empty. Are you sure you're using the correct report name")
    if not (report_url or url_name):
        raise ValueError("report_url or url_name must be provided")

    # if not report.chart_settings:
    kwargs.setdefault("display_chart", bool(report.chart_settings))
    kwargs.setdefault("display_table", True)

    kwargs.setdefault("display_chart_selector", kwargs["display_chart"])
    kwargs.setdefault("display_title", True)

    passed_title = kwargs.get("title", None)
    kwargs["title"] = passed_title or report.get_report_title()
    kwargs["report_url"] = report_url
    if not report_url:
        kwargs["report_url"] = reverse(url_name)

    kwargs.setdefault("extra_params", "")

    template = get_template(template_name or "slick_reporting/widget_template.html")

    return template.render(context=kwargs)


@register.simple_tag
def add_jquery():
    if SLICK_REPORTING_JQUERY_URL:
        return mark_safe(f'<script src="{SLICK_REPORTING_JQUERY_URL}"></script>')
    return ""


@register.simple_tag
def get_charts_media(chart_settings):
    charts_dict = SLICK_REPORTING_SETTINGS["CHARTS"]
    media = Media()
    if chart_settings == "all":
        available_types = charts_dict.keys()
    else:
        available_types = [chart["engine_name"] for chart in chart_settings]
        available_types = set(available_types)

    for type in available_types:
        media += Media(css=charts_dict.get(type, {}).get("css", {}), js=charts_dict.get(type, {}).get("js", []))
    return media


@register.simple_tag
def get_slick_reporting_media():
    from django.forms import Media

    media = get_media()
    return Media(css=media["css"], js=media["js"])


@register.simple_tag
def get_slick_reporting_settings():
    return dict(SLICK_REPORTING_SETTINGS)



from django import apps


class ReportAppConfig(apps.AppConfig):
    verbose_name = "Slick Reporting"
    name = "slick_reporting"

    def ready(self):
        super().ready()
        from . import fields  # noqa



from django.conf import settings
from django.utils.functional import lazy
from django.utils.translation import gettext_lazy as _

import datetime


def get_first_of_this_year():
    d = datetime.datetime.today()
    return datetime.datetime(d.year, 1, 1, 0, 0)


def get_end_of_this_year():
    d = datetime.datetime.today()
    return datetime.datetime(d.year + 1, 1, 1, 0, 0)


def get_start_date():
    start_date = getattr(settings, "SLICK_REPORTING_DEFAULT_START_DATE", False)
    return start_date or get_first_of_this_year()


def get_end_date():
    end_date = getattr(settings, "SLICK_REPORTING_DEFAULT_END_DATE", False)
    return end_date or datetime.datetime.today()


SLICK_REPORTING_DEFAULT_START_DATE = lazy(get_start_date, datetime.datetime)()
SLICK_REPORTING_DEFAULT_END_DATE = lazy(get_end_date, datetime.datetime)()


SLICK_REPORTING_DEFAULT_CHARTS_ENGINE = getattr(settings, "SLICK_REPORTING_DEFAULT_CHARTS_ENGINE", "highcharts")


SLICK_REPORTING_JQUERY_URL = getattr(
    settings, "SLICK_REPORTING_JQUERY_URL", "https://code.jquery.com/jquery-3.7.0.min.js"
)


SLICK_REPORTING_SETTINGS_DEFAULT = {
    "JQUERY_URL": SLICK_REPORTING_JQUERY_URL,
    "DEFAULT_START_DATE_TIME": get_start_date(),
    "DEFAULT_END_DATE_TIME": get_end_date(),
    "DEFAULT_CHARTS_ENGINE": SLICK_REPORTING_DEFAULT_CHARTS_ENGINE,
    "MEDIA": {
        "override": False,
        "js": (
            "https://cdn.jsdelivr.net/momentjs/latest/moment.min.js",
            "https://cdn.jsdelivr.net/npm/select2@4.1.0-rc.0/dist/js/select2.min.js",
            "https://cdn.datatables.net/1.13.4/js/jquery.dataTables.min.js",
            "https://cdn.datatables.net/1.13.4/js/dataTables.bootstrap5.min.js",
            "slick_reporting/slick_reporting.js",
            "slick_reporting/slick_reporting.report_loader.js",
            "slick_reporting/slick_reporting.datatable.js",
        ),
        "css": {
            "all": (
                "https://cdn.jsdelivr.net/npm/select2@4.1.0-rc.0/dist/css/select2.min.css",
                "https://cdn.datatables.net/1.13.4/css/dataTables.bootstrap5.min.css",
            )
        },
    },
    "FONT_AWESOME": {
        "CSS_URL": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/5.15.3/css/all.min.css",
        "ICONS": {
            "pie": "fas fa-chart-pie",
            "bar": "fas fa-chart-bar",
            "line": "fas fa-chart-line",
            "area": "fas fa-chart-area",
            "column": "fas fa-chart-column",
        },
    },
    "CHARTS": {
        "highcharts": {
            "entryPoint": "$.slick_reporting.highcharts.displayChart",
            "js": ("https://code.highcharts.com/highcharts.js", "slick_reporting/slick_reporting.highchart.js"),
        },
        "chartsjs": {
            "entryPoint": "$.slick_reporting.chartsjs.displayChart",
            "js": ("https://cdn.jsdelivr.net/npm/chart.js", "slick_reporting/slick_reporting.chartsjs.js"),
        },
    },
    "MESSAGES": {
        "total": _("Total"),
        "export_to_csv": _("Export to CSV"),
    },
}


def get_slick_reporting_settings():
    slick_settings = SLICK_REPORTING_SETTINGS_DEFAULT.copy()
    slick_chart_settings = slick_settings["CHARTS"].copy()

    user_settings = getattr(settings, "SLICK_REPORTING_SETTINGS", {})
    user_chart_settings = user_settings.get("CHARTS", {})

    slick_chart_settings.update(user_chart_settings)
    slick_settings.update(user_settings)
    slick_settings["CHARTS"] = slick_chart_settings

    # slick_settings = {**SLICK_REPORTING_SETTINGS_DEFAULT, **getattr(settings, "SLICK_REPORTING_SETTINGS", {})}
    start_date = getattr(settings, "SLICK_REPORTING_DEFAULT_START_DATE", False)
    end_date = getattr(settings, "SLICK_REPORTING_DEFAULT_END_DATE", False)
    # backward compatibility, todo remove in next major release
    if start_date:
        slick_settings["DEFAULT_START_DATE_TIME"] = start_date
    if end_date:
        slick_settings["DEFAULT_END_DATE_TIME"] = end_date

    return slick_settings


SLICK_REPORTING_SETTINGS = lazy(get_slick_reporting_settings, dict)()


def get_media():
    return SLICK_REPORTING_SETTINGS["MEDIA"]



from collections import OrderedDict

from crispy_forms.helper import FormHelper
from django import forms
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from . import app_settings
from .helpers import get_foreign_keys, get_field_from_query_text

TIME_SERIES_CHOICES = (
    ("monthly", _("Monthly")),
    ("weekly", _("Weekly")),
    ("annually", _("Yearly")),
    ("daily", _("Daily")),
)


def default_formfield_callback(f, **kwargs):
    kwargs["required"] = False
    kwargs["help_text"] = ""
    return f.formfield(**kwargs)


def get_crispy_helper(
    foreign_keys_map=None,
    crosstab_model=None,
    crosstab_key_name=None,
    crosstab_display_compute_remainder=False,
    add_date_range=True,
):
    from crispy_forms.helper import FormHelper
    from crispy_forms.layout import Column, Layout, Div, Row, Field

    foreign_keys_map = foreign_keys_map or []
    helper = FormHelper()
    helper.form_class = "form-horizontal"
    helper.label_class = "col-sm-2 col-md-2 col-lg-2"
    helper.field_class = "col-sm-10 col-md-10 col-lg-10"
    helper.form_tag = False
    helper.disable_csrf = True
    helper.render_unmentioned_fields = True

    helper.layout = Layout()
    if add_date_range:
        helper.layout.fields.append(
            Row(
                Column(Field("start_date"), css_class="col-sm-6"),
                Column(Field("end_date"), css_class="col-sm-6"),
                css_class="raReportDateRange",
            ),
        )
    filters_container = Div(css_class="mt-20", style="margin-top:20px")
    # first add the crosstab model and its display reimder then the rest of the fields
    if crosstab_model:
        filters_container.append(Field(crosstab_key_name))
        if crosstab_display_compute_remainder:
            filters_container.append(Field("crosstab_compute_remainder"))

    for k in foreign_keys_map:
        if k != crosstab_key_name:
            filters_container.append(Field(k))
    helper.layout.fields.append(filters_container)

    return helper


def get_choices_form_queryset_list(qs):
    choices = []
    for row in qs:
        choices.append((row, row))
    return choices


class OrderByForm(forms.Form):
    order_by = forms.CharField(required=False)

    def get_order_by(self, default_field=None):
        """
        Get the order by specified by teh form or the default field if provided
        :param default_field:
        :return: tuple of field and direction
        """
        if self.is_valid():
            order_field = self.cleaned_data["order_by"]
            order_field = order_field or default_field
            if order_field:
                return self.parse_order_by_field(order_field)
        return None, None

    def parse_order_by_field(self, order_field):
        """
        Specify the field and direction
        :param order_field: the field to order by
        :return: tuple of field and direction
        """
        if order_field:
            asc = True
            if order_field[0:1] == "-":
                order_field = order_field[1:]
                asc = False
            return order_field, not asc
        return None, None


class BaseReportForm:
    def get_filters(self):
        raise NotImplementedError(
            "get_filters() must be implemented in subclass,"
            "should return a tuple of (Q objects, kwargs filter) to be passed to QuerySet.filter()"
        )

    def get_start_date(self):
        raise NotImplementedError("get_start_date() must be implemented in subclass," "should return a datetime object")

    def get_end_date(self):
        raise NotImplementedError("get_end_date() must be implemented in subclass," "should return a datetime object")

    def get_crosstab_compute_remainder(self):
        raise NotImplementedError(
            "get_crosstab_compute_remainder() must be implemented in subclass," "should return a boolean value"
        )

    def get_crosstab_ids(self):
        raise NotImplementedError(
            "get_crosstab_ids() must be implemented in subclass," "should return a list of ids to be used for crosstab"
        )

    def get_time_series_pattern(self):
        raise NotImplementedError(
            "get_time_series_pattern() must be implemented in subclass,"
            "should return a string value of a valid time series pattern"
        )

    def get_crispy_helper(self):
        # return a default helper
        helper = FormHelper()
        helper.form_class = "form-horizontal"
        helper.label_class = "col-sm-2 col-md-2 col-lg-2"
        helper.field_class = "col-sm-10 col-md-10 col-lg-10"
        helper.form_tag = False
        helper.disable_csrf = True
        helper.render_unmentioned_fields = True
        return helper


class SlickReportForm(BaseReportForm):
    """
    Holds basic function
    """

    def get_start_date(self):
        return self.cleaned_data.get("start_date")

    def get_end_date(self):
        return self.cleaned_data.get("end_date")

    def get_time_series_pattern(self):
        return self.cleaned_data.get("time_series_pattern")

    def get_filters(self):
        """
        Get the foreign key filters for report queryset, excluding crosstab ids, handled by `get_crosstab_ids()`
        :return: a dicttionary of filters to be used with QuerySet.filter(**returned_value)
        """
        _values = {}
        if self.is_valid():
            fk_keys = getattr(self, "foreign_keys", [])
            if fk_keys:
                fk_keys = fk_keys.items()
            for key, field in fk_keys:
                if key in self.cleaned_data and not key == self.crosstab_key_name:
                    val = self.cleaned_data[key]
                    if val:
                        val = [x for x in val.values_list("pk", flat=True)]
                        _values["%s__in" % key] = val
            return None, _values

    @cached_property
    def crosstab_key_name(self):
        # todo get the model more accurately
        """
        return the actual foreignkey field name by simply adding an '_id' at the end.
        This is hook is to customize this naieve approach.
        :return: key: a string that should be in self.cleaned_data
        """
        if self.crosstab_field_klass:
            return self.crosstab_field_klass.attname
        return f"{self.crosstab_model}_id"

    def get_crosstab_ids(self):
        """
        Get the crosstab ids so they can be sent to the report generator.
        :return:
        """
        if self.crosstab_field_klass:
            if self.crosstab_field_klass.is_relation:
                qs = self.cleaned_data.get(self.crosstab_key_name)
                return [x for x in qs.values_list(self.crosstab_field_related_name, flat=True)]
            else:
                return self.cleaned_data.get(self.crosstab_key_name)
        return []

    def get_crosstab_compute_remainder(self):
        return self.cleaned_data.get("crosstab_compute_remainder", True)

    def get_crispy_helper(self, foreign_keys_map=None, crosstab_model=None, **kwargs):
        return get_crispy_helper(
            self.foreign_keys,
            crosstab_model=getattr(self, "crosstab_model", None),
            crosstab_key_name=getattr(self, "crosstab_key_name", None),
            crosstab_display_compute_remainder=getattr(self, "crosstab_display_compute_remainder", False),
            **kwargs,
        )


def _default_foreign_key_widget(f_field):
    return {
        "form_class": forms.ModelMultipleChoiceField,
        "required": False,
    }


def report_form_factory(
    model,
    crosstab_model=None,
    display_compute_remainder=True,
    fkeys_filter_func=None,
    foreign_key_widget_func=None,
    excluded_fields=None,
    initial=None,
    required=None,
    show_time_series_selector=False,
    time_series_selector_choices=None,
    time_series_selector_default="",
    time_series_selector_label=None,
    time_series_selector_allow_empty=False,
):
    """
    Create a Report Form based on the report_model passed by
    1. adding a start_date and end_date fields
    2. extract all ForeignKeys on the report_model

    :param model: the report_model
    :param crosstab_model: crosstab model if any
    :param display_compute_remainder:  relevant only if crosstab_model is specified. Control if we show the check to
    display the rest.
    :param fkeys_filter_func:  a receives an OrderedDict of Foreign Keys names and their model field instances found on
           the model, return the OrderedDict that would be used
    :param foreign_key_widget_func: receives a Field class return the used widget like this
           {'form_class': forms.ModelMultipleChoiceField, 'required': False, }
    :param excluded_fields: a list of fields to be excluded from the report form
    :param initial a dict for fields initial
    :param required a list of fields that should be marked as required
    :return:
    """
    crosstab_field_related_name = ""
    crosstab_field_klass = None
    foreign_key_widget_func = foreign_key_widget_func or _default_foreign_key_widget
    fkeys_filter_func = fkeys_filter_func or (lambda x: x)

    # gather foreign keys
    initial = initial or {}
    required = required or []
    fkeys_map = get_foreign_keys(model)
    excluded_fields = excluded_fields or []
    for excluded in excluded_fields:
        del fkeys_map[excluded]

    fkeys_map = fkeys_filter_func(fkeys_map)

    fkeys_list = []
    fields = OrderedDict()

    fields["start_date"] = forms.DateTimeField(
        required=False,
        label=_("From date"),
        initial=initial.get("start_date", "") or app_settings.SLICK_REPORTING_SETTINGS["DEFAULT_START_DATE_TIME"],
        widget=forms.DateTimeInput(attrs={"autocomplete": "off"}),
    )

    fields["end_date"] = forms.DateTimeField(
        required=False,
        label=_("To  date"),
        initial=initial.get("end_date", "") or app_settings.SLICK_REPORTING_SETTINGS["DEFAULT_END_DATE_TIME"],
        widget=forms.DateTimeInput(attrs={"autocomplete": "off"}),
    )

    if show_time_series_selector:
        time_series_choices = list(TIME_SERIES_CHOICES)
        if time_series_selector_allow_empty:
            time_series_choices.insert(0, ("", "---------"))

        fields["time_series_pattern"] = forms.ChoiceField(
            required=False,
            initial=time_series_selector_default,
            label=time_series_selector_label or _("Period Pattern"),
            choices=time_series_selector_choices or TIME_SERIES_CHOICES,
        )

    for name, f_field in fkeys_map.items():
        fkeys_list.append(name)
        field_attrs = foreign_key_widget_func(f_field)
        if name in required:
            field_attrs["required"] = True
        field_attrs["initial"] = initial.get(name, "")
        fields[name] = f_field.formfield(**field_attrs)

    if crosstab_model:
        # todo Enhance, add tests , cover cases
        # Crosstab is a foreign key on model
        # crosstab is a Char field on model
        # crosstab is a traversing fk field
        # crosstab is a traversing Char / choice field

        if display_compute_remainder:
            fields["crosstab_compute_remainder"] = forms.BooleanField(
                required=False, label=_("Display the crosstab remainder"), initial=True
            )

        crosstab_field_klass = get_field_from_query_text(crosstab_model, model)
        if crosstab_field_klass.is_relation:
            crosstab_field_related_name = crosstab_field_klass.to_fields[0]
        else:
            crosstab_field_related_name = crosstab_field_klass.name

        if "__" in crosstab_model:  # traversing field, it won't be added naturally to the form
            if crosstab_field_klass.is_relation:
                pass
            else:
                fields[crosstab_field_related_name] = forms.MultipleChoiceField(
                    choices=get_choices_form_queryset_list(
                        list(
                            crosstab_field_klass.model.objects.values_list(
                                crosstab_field_related_name, flat=True
                            ).distinct()
                        )
                    ),
                    required=False,
                    label=crosstab_field_klass.verbose_name,
                )

    bases = (
        SlickReportForm,
        forms.BaseForm,
    )
    new_form = type(
        "ReportForm",
        bases,
        {
            "base_fields": fields,
            "_fkeys": fkeys_list,
            "foreign_keys": fkeys_map,
            "crosstab_model": crosstab_model,
            "crosstab_display_compute_remainder": display_compute_remainder,
            "crosstab_field_related_name": crosstab_field_related_name,
            "crosstab_field_klass": crosstab_field_klass,
        },
    )
    return new_form



from __future__ import unicode_literals

from django.contrib.admin.sites import AlreadyRegistered, NotRegistered


class ReportFieldRegistry(object):
    def __init__(self):
        super(ReportFieldRegistry, self).__init__()
        self._registry = {}  # holds

    def register(self, report_field, override=False):
        """
        Register a report_field into the registry,
        :param report_field:
        :param override: if True, a report_field will get replaced if found, else it would throw an AlreadyRegistered
        :return: report_field passed
        """
        if report_field.name in self._registry and not override:
            raise AlreadyRegistered(f"The field name {report_field.name} is used before and `override` is False")

        self._registry[report_field.name] = report_field
        return report_field

    def unregister(self, report_field):
        """
        To unregister a Report Field
        :param report_field: a Report field class or a ReportField Name
        :return: None
        """
        name = report_field if isinstance(report_field, str) else report_field.name
        if name not in self._registry:
            raise NotRegistered(report_field)
        del self._registry[name]

    def get_field_by_name(self, name):
        if name in self._registry:
            return self._registry[name]
        else:
            raise KeyError(
                f'{name} is not found in the report field registry. Options are {",".join(self.get_all_report_fields_names())}'
            )

    def get_all_report_fields_names(self):
        return list(self._registry.keys())


field_registry = ReportFieldRegistry()



import datetime
import logging
from dataclasses import dataclass
from inspect import isclass

from django.core.exceptions import ImproperlyConfigured, FieldDoesNotExist
from django.db.models import Q, ForeignKey

from .app_settings import SLICK_REPORTING_DEFAULT_CHARTS_ENGINE
from .fields import ComputationField
from .helpers import get_field_from_query_text
from .registry import field_registry
from . import app_settings

logger = logging.getLogger(__name__)


@dataclass
class Chart:
    title: str
    type: str
    data_source: list
    title_source: list
    plot_total: bool = False
    engine: str = ""
    entryPoint: str = ""
    COLUMN = "column"
    LINE = "line"
    PIE = "pie"
    BAR = "bar"
    AREA = "area"

    def to_dict(self):
        return dict(
            title=self.title,
            type=self.type,
            data_source=self.data_source,
            title_source=self.title_source,
            plot_total=self.plot_total,
            engine=self.engine,
            entryPoint=self.entryPoint,
        )


class ReportGeneratorAPI:
    report_model = None
    """The main model where data is """

    queryset = None
    """If set, the report will use this queryset instead of the report_model"""

    """
    Class to generate a Json Object containing report data.
    """
    date_field = ""
    """Main date field to use whenever date filter is needed"""

    start_date_field_name = None
    """If set, the report will use this field to filter the start date, default to date_field"""

    end_date_field_name = None
    """If set, the report will use this field to filter the end date, default to date_field"""

    print_flag = None

    group_by = None
    """The field to use for grouping, if not set then the report is expected to be a sub version of the report model"""

    group_by_custom_querysets = None
    """A List of querysets representing different group by options"""
    group_by_custom_querysets_column_verbose_name = None

    columns = None
    """A list of column names.
    Columns names can be 

    1. A Computation Field

    2. If group_by is set, then any field on the group_by model

    3. If group_by is not set, then any field name on the report_model / queryset

    4. A callable on the generator

    5. Special __time_series__, and __crosstab__ 
       Those can be use to control the position of the time series inside the columns, defaults it's appended at the end

       Example:
       columns = ['product_id', '__time_series__', 'col_b']
       Same is true with __crosstab__ 

    You can customize aspects of the column by adding it as a tuple like this 
        ('field_name', dict(verbose_name=_('My Enhanced Verbose_name'))


     """

    time_series_pattern = ""
    """
    If set the Report will compute a time series.

    Possible options are: daily, weekly, semimonthly, monthly, quarterly, semiannually, annually and custom.

    if `custom` is set, you'd need to override  `get_custom_time_series_dates`
    """
    time_series_columns = None
    """
    a list of Calculation Field names which will be included in the series calculation.
     Example: ['__total__', '__total_quantity__'] with compute those 2 fields for all the series

    """

    time_series_custom_dates = None
    """
    Used with `time_series_pattern` set to 'custom'
    It's a list of tuple, each tuple represent start date & end date
    Example: [ (start_date_1, end_date_1), (start_date_2, end_date_2), ....]
    """

    crosstab_model = None  # deprecated

    crosstab_field = None
    """
    If set, a cross tab over this model selected ids (via `crosstab_ids`)  
    """

    crosstab_columns = None
    """The computation fields which will be computed for each crosstab-ed ids """

    crosstab_ids = None
    """A list is the ids to create a crosstab report on"""

    crosstab_ids_custom_filters = None

    crosstab_compute_remainder = True
    """Include an an extra crosstab_columns for the outer group ( ie: all expects those `crosstab_ids`) """

    limit_records = None
    """Serves are a main limit to  the returned data of the report_model.
    Can be beneficial if the results may be huge.
    """
    swap_sign = False


class ReportGenerator(ReportGeneratorAPI, object):
    """
    The main class responsible generating the report and managing the flow
    """

    field_registry_class = field_registry
    """You can have a custom computation field locator! It only needs a `get_field_by_name(string)` 
    and returns a ReportField`"""

    def __init__(
        self,
        report_model=None,
        main_queryset=None,
        start_date=None,
        end_date=None,
        date_field=None,
        q_filters=None,
        kwargs_filters=None,
        group_by=None,
        group_by_custom_querysets=None,
        group_by_custom_querysets_column_verbose_name=None,
        columns=None,
        time_series_pattern=None,
        time_series_columns=None,
        time_series_custom_dates=None,
        crosstab_field=None,
        crosstab_columns=None,
        crosstab_ids=None,
        crosstab_ids_custom_filters=None,
        crosstab_compute_remainder=None,
        swap_sign=False,
        show_empty_records=None,
        print_flag=False,
        doc_type_plus_list=None,
        doc_type_minus_list=None,
        limit_records=False,
        format_row_func=None,
        container_class=None,
        start_date_field_name=None,
        end_date_field_name=None,
    ):
        """

        :param report_model: Main model containing the data
        :param main_queryset: Default to report_model.objects
        :param start_date:
        :param end_date:
        :param date_field:
        :param q_filters:
        :param kwargs_filters:
        :param group_by:
        :param columns:
        :param time_series_pattern:
        :param time_series_columns:
        :param crosstab_model:
        :param crosstab_columns:
        :param crosstab_ids:
        :param crosstab_compute_remainder:
        :param swap_sign:
        :param show_empty_records:
        :param base_model:
        :param print_flag:
        :param doc_type_plus_list:
        :param doc_type_minus_list:
        :param limit_records:
        """
        from .app_settings import (
            SLICK_REPORTING_DEFAULT_START_DATE,
            SLICK_REPORTING_DEFAULT_END_DATE,
        )

        super().__init__()

        self.report_model = self.report_model or report_model
        if self.queryset is None:
            self.queryset = main_queryset

        if not self.report_model and self.queryset is None:
            raise ImproperlyConfigured("report_model or queryset must be set on a class level or via init")

        main_queryset = self.report_model.objects if self.queryset is None else self.queryset

        self.start_date = start_date or datetime.datetime.combine(
            SLICK_REPORTING_DEFAULT_START_DATE.date(),
            SLICK_REPORTING_DEFAULT_START_DATE.time(),
        )

        self.end_date = end_date or datetime.datetime.combine(
            SLICK_REPORTING_DEFAULT_END_DATE.date(),
            SLICK_REPORTING_DEFAULT_END_DATE.time(),
        )
        self.date_field = self.date_field or date_field

        self.start_date_field_name = self.start_date_field_name or start_date_field_name or self.date_field
        self.end_date_field_name = self.end_date_field_name or end_date_field_name or self.date_field

        self.q_filters = q_filters or []
        self.kwargs_filters = kwargs_filters or {}
        self.crosstab_field = self.crosstab_field or crosstab_field

        self.crosstab_columns = crosstab_columns or self.crosstab_columns or []
        self.crosstab_ids = self.crosstab_ids or crosstab_ids or []
        self.crosstab_ids_custom_filters = self.crosstab_ids_custom_filters or crosstab_ids_custom_filters or []

        self.crosstab_compute_remainder = (
            self.crosstab_compute_remainder if crosstab_compute_remainder is None else crosstab_compute_remainder
        )

        self.format_row = format_row_func or self._default_format_row

        main_queryset = self.report_model.objects if main_queryset is None else main_queryset
        # todo revise & move somewhere nicer, List Report need to override the resetting of order
        main_queryset = self._remove_order(main_queryset)

        self.columns = columns or self.columns or []
        self.group_by = group_by or self.group_by

        self.group_by_custom_querysets = group_by_custom_querysets or self.group_by_custom_querysets or []

        self.group_by_custom_querysets_column_verbose_name = (
            group_by_custom_querysets_column_verbose_name or self.group_by_custom_querysets_column_verbose_name or ""
        )
        self.time_series_pattern = self.time_series_pattern or time_series_pattern
        self.time_series_columns = self.time_series_columns or time_series_columns
        self.time_series_custom_dates = self.time_series_custom_dates or time_series_custom_dates
        self.container_class = container_class

        if not (self.date_field or (self.start_date_field_name and self.end_date_field_name)) and (
            self.time_series_pattern or self.crosstab_field or self.group_by
        ):
            raise ImproperlyConfigured(
                f"date_field or [start_date_field_name and end_date_field_name] must "
                f"be set for {container_class or self}"
            )

        self._prepared_results = {}
        self.report_fields_classes = {}

        self._report_fields_dependencies = {
            "time_series": {},
            "crosstab": {},
            "normal": {},
        }
        self.existing_dependencies = {"series": [], "matrix": [], "normal": []}

        self.print_flag = print_flag or self.print_flag

        # todo validate columns is not empty (if no time series / cross tab)

        if self.group_by:
            try:
                self.group_by_field = get_field_from_query_text(self.group_by, self.report_model)

            except (IndexError, AttributeError):
                raise ImproperlyConfigured(
                    f"Can not find group_by field:{self.group_by} in report_model {self.report_model} "
                )
            if "__" not in self.group_by:
                self.group_by_field_attname = self.group_by_field.attname
            else:
                self.group_by_field_attname = self.group_by

        else:
            self.group_by_field_attname = None

        # doc_types = form.get_doc_type_plus_minus_lists()
        doc_types = [], []
        self.doc_type_plus_list = list(doc_type_plus_list) if doc_type_plus_list else doc_types[0]
        self.doc_type_minus_list = list(doc_type_minus_list) if doc_type_minus_list else doc_types[1]

        self.swap_sign = self.swap_sign or swap_sign
        self.limit_records = self.limit_records or limit_records

        # todo delete this
        self.show_empty_records = False  # show_empty_records if show_empty_records else self.show_empty_records

        # Preparing actions
        self._parse()

        self.main_queryset = self.prepare_queryset(main_queryset)
        self._prepare_report_dependencies()

    def prepare_queryset(self, queryset):
        if self.group_by_custom_querysets:
            return [{"__index__": i} for i, v in enumerate(self.group_by_custom_querysets)]
        elif self.group_by:
            main_queryset = self._apply_queryset_options(queryset)

            if type(self.group_by_field) is ForeignKey:
                ids = main_queryset.values_list(self.group_by_field_attname).distinct()
                # uses the same logic that is in Django's query.py when fields is empty in values() call
                concrete_fields = [f.name for f in self.group_by_field.related_model._meta.concrete_fields]
                # add database columns that are not already in concrete_fields
                final_fields = concrete_fields + list(set(self.get_database_columns()) - set(concrete_fields))
                return self.group_by_field.related_model.objects.filter(
                    **{f"{self.group_by_field.target_field.name}__in": ids}
                ).values(*final_fields)
            else:
                return main_queryset.distinct().values(self.group_by_field_attname)

        return [{}]

    def _remove_order(self, main_queryset):
        """
        Remove order_by from the main queryset
        :param main_queryset:
        :return:
        """
        # if main_queryset.query.order_by:
        main_queryset = main_queryset.order_by()
        return main_queryset

    def _apply_queryset_options(self, query, fields=None):
        """
        Apply the filters to the main queryset which will computed results be mapped to
        :param query:
        :param fields:
        :return:
        """
        filters = {}
        if self.date_field:
            filters = {
                f"{self.start_date_field_name}__gt": self.start_date,
                f"{self.end_date_field_name}__lte": self.end_date,
            }
        filters.update(self.kwargs_filters)

        if filters:
            query = query.filter(**filters)
        if self.q_filters:
            query = query.filter(*self.q_filters)
        if fields:
            return query.values(*fields)
        return query.values()

    def _construct_crosstab_filter(self, col_data, queryset_filters=None):
        """
        In charge of adding the needed crosstab filter, specific to the case of is_remainder or not
        :param col_data:
        :return:
        """
        if queryset_filters:
            return queryset_filters[0], queryset_filters[1]

        if "__" in col_data["crosstab_field"]:
            column_name = col_data["crosstab_field"]
        else:
            field = get_field_from_query_text(col_data["crosstab_field"], self.report_model)
            column_name = field.column
        if col_data["is_remainder"] and not queryset_filters:
            filters = [~Q(**{f"{column_name}__in": self.crosstab_ids})]
        else:
            filters = [Q(**{f"{column_name}": col_data["id"]})]
        return filters, {}

    def _prepare_report_dependencies(self):
        from .fields import ComputationField

        all_columns = (
            ("normal", self._parsed_columns),
            ("time_series", self._time_series_parsed_columns),
            ("crosstab", self._crosstab_parsed_columns),
        )
        for window, window_cols in all_columns:
            for col_data in window_cols:
                klass = col_data["ref"]

                if isclass(klass) and issubclass(klass, ComputationField):
                    dependencies_names = klass.get_full_dependency_list()

                    # check if any of these dependencies is on the report, if found we call the child to
                    # resolve the value for its parent avoiding extra database call
                    fields_on_report = [
                        x
                        for x in window_cols
                        if x["ref"] in dependencies_names
                        and (
                            (
                                window == "time_series"
                                and x.get("start_date", "") == col_data.get("start_date", "")
                                and x.get("end_date") == col_data.get("end_date")
                            )
                            or window == "crosstab"
                            and x.get("id") == col_data.get("id")
                        )
                    ]
                    for field in fields_on_report:
                        self._report_fields_dependencies[window][field["name"]] = col_data["name"]
            for col_data in window_cols:
                klass = col_data["ref"]
                name = col_data["name"]

                # if column has a dependency then skip it
                if not (isclass(klass) and issubclass(klass, ComputationField)):
                    continue
                if self._report_fields_dependencies[window].get(name, False):
                    continue

                report_class = klass(
                    self.doc_type_plus_list,
                    self.doc_type_minus_list,
                    group_by=self.group_by,
                    report_model=self.report_model,
                    date_field=self.date_field,
                    queryset=self.queryset,
                    group_by_custom_querysets=self.group_by_custom_querysets,
                )

                q_filters = None
                date_filter = {
                    f"{self.start_date_field_name}__gte": col_data.get("start_date", self.start_date),
                    f"{self.end_date_field_name}__lt": col_data.get("end_date", self.end_date),
                }
                date_filter.update(self.kwargs_filters)
                if window == "crosstab" or col_data.get("computation_flag", "") == "crosstab":
                    q_filters, kw_filters = col_data["queryset_filters"]
                    date_filter.update(kw_filters)

                report_class.init_preparation(q_filters, date_filter)
                self.report_fields_classes[name] = report_class

    # @staticmethod
    def get_primary_key_name(self, model):
        if self.group_by_custom_querysets:
            return "__index__"
        for field in model._meta.fields:
            if field.primary_key:
                return field.attname
        return ""

    def _get_record_data(self, obj, columns):
        """
        the function is run for every obj in the main_queryset
        :param obj: current row
        :param: columns： The columns we iterate on
        :return: a dict object containing all needed data
        """

        data = {}
        group_by_val = None
        if self.group_by_custom_querysets:
            group_by_val = str(obj["__index__"])

        elif self.group_by:
            if self.group_by_field.related_model and "__" not in self.group_by:
                primary_key_name = self.get_primary_key_name(self.group_by_field.related_model)
            else:
                primary_key_name = self.group_by_field_attname

            column_data = obj.get(primary_key_name, obj.get("id"))
            group_by_val = str(column_data)

        for window, window_cols in columns:
            for col_data in window_cols:
                name = col_data["name"]

                if col_data.get("source", "") == "attribute_field":
                    data[name] = col_data["ref"](obj, data)
                elif col_data.get("source", "") == "container_class_attribute_field":
                    data[name] = col_data["ref"](obj, data)

                elif (
                    col_data.get("source", "") == "magic_field" and (self.group_by or self.group_by_custom_querysets)
                ) or (not (self.group_by or self.group_by_custom_querysets)):
                    source = self._report_fields_dependencies[window].get(name, False)

                    if source:
                        computation_class = self.report_fields_classes[source]
                        # the computation field is being asked from another computation field that requires it.
                        value = computation_class.get_dependency_value(group_by_val, col_data["ref"].name)
                    else:
                        try:
                            computation_class = self.report_fields_classes[name]
                        except KeyError:
                            continue
                        value = computation_class.do_resolve(group_by_val, data)
                    if self.swap_sign:
                        value = -value
                    data[name] = value

                else:
                    data[name] = obj.get(name, "")
        return data

    def get_report_data(self):
        main_queryset = self.main_queryset[: self.limit_records] if self.limit_records else self.main_queryset

        all_columns = (
            ("normal", self._parsed_columns),
            ("time_series", self._time_series_parsed_columns),
            ("crosstab", self._crosstab_parsed_columns),
        )

        get_record_data = self._get_record_data
        format_row = self.format_row
        data = [format_row(get_record_data(obj, all_columns)) for obj in main_queryset]
        return data

    def _default_format_row(self, row_obj):
        """
        Hook where you can format row values like properly format a date
        :param row_obj:
        :return:
        """
        return row_obj

    @staticmethod
    def check_columns(
        cls,
        columns,
        group_by,
        report_model,
        container_class=None,
        group_by_custom_querysets=None,
    ):
        """
        Check and parse the columns, throw errors in case an item in the columns cant not identified
        :param columns: List of columns
        :param group_by: group by field if any
        :param report_model: the report model
        :param container_class: a class to search for custom columns attribute in, typically the ReportView
        :param group_by_custom_querysets a list of group by custom queries Or None.
        :return: List of dict, each dict contains relevant data to the respective field in `columns`
        """

        group_by_model = None
        if group_by_custom_querysets:
            if "__index__" not in columns:
                columns.insert(0, "__index__")

        if group_by:
            try:
                group_by_field = [x for x in report_model._meta.get_fields() if x.name == group_by.split("__")[0]][0]
            except IndexError:
                raise ImproperlyConfigured(
                    f"ReportView {cls}: Could not find the group_by field: `{group_by}` in "
                    f"report_model: `{report_model}`"
                )
            if group_by_field.is_relation:
                group_by_model = group_by_field.related_model
            else:
                group_by_model = report_model

        parsed_columns = []
        for col in columns:
            options = {}
            if type(col) is tuple:
                col, options = col

            if col in ["__time_series__", "__crosstab__"]:
                #     These are placeholder not real computation field
                continue

            magic_field_class = None
            attribute_field = None
            is_container_class_attribute = False

            if isinstance(col, str):
                attribute_field = getattr(cls, col, None)
                if attribute_field is None:
                    is_container_class_attribute = True
                    attribute_field = getattr(container_class, col, None)

            elif issubclass(col, ComputationField):
                magic_field_class = col

            try:
                magic_field_class = magic_field_class or field_registry.get_field_by_name(col)
            except KeyError:
                magic_field_class = None

            if attribute_field:
                col_data = {
                    "name": col,
                    "verbose_name": getattr(attribute_field, "verbose_name", col),
                    "source": "container_class_attribute_field" if is_container_class_attribute else "attribute_field",
                    "ref": attribute_field,
                    "type": "text",
                }
            elif magic_field_class:
                # a magic field
                col_data = {
                    "name": magic_field_class.name,
                    "verbose_name": magic_field_class.verbose_name,
                    "source": "magic_field",
                    "ref": magic_field_class,
                    "type": magic_field_class.type,
                    "is_summable": magic_field_class.is_summable,
                }
            else:
                # A database field
                if group_by_custom_querysets and col == "__index__":
                    # group by custom queryset special case: which is the index
                    col_data = {
                        "name": col,
                        "verbose_name": cls.group_by_custom_querysets_column_verbose_name,
                        "source": "database",
                        "ref": "",
                        "type": "text",
                    }
                    col_data.update(options)
                    parsed_columns.append(col_data)
                    continue

                model_to_use = group_by_model if group_by and "__" not in group_by else report_model
                group_by_str = str(group_by)
                if "__" in group_by_str:
                    related_model = get_field_from_query_text(group_by, model_to_use).related_model
                    model_to_use = related_model if related_model else model_to_use

                try:
                    if "__" in col:
                        # A traversing link order__client__email
                        field = get_field_from_query_text(col, model_to_use)
                    else:
                        field = model_to_use._meta.get_field(col)
                except FieldDoesNotExist:
                    field = getattr(container_class, col, False)

                    if not field:
                        raise FieldDoesNotExist(
                            f'Field "{col}" not found either as an attribute to the generator class {cls}, '
                            f'{f"Container class {container_class}," if container_class else ""}'
                            f'or a computation field, or a database column for the model "{model_to_use}"'
                        )

                col_data = {
                    "name": col,
                    "verbose_name": getattr(field, "verbose_name", col),
                    "source": "database",
                    "ref": field,
                    "type": "choice" if field.choices else field.get_internal_type(),
                }
            col_data.update(options)
            parsed_columns.append(col_data)
        return parsed_columns

    def _parse(self):
        self.parsed_columns = self.check_columns(
            self,
            self.columns,
            self.group_by,
            self.report_model,
            self.container_class,
            self.group_by_custom_querysets,
        )
        self._parsed_columns = list(self.parsed_columns)
        self._crosstab_parsed_columns = self.get_crosstab_parsed_columns()
        self._time_series_parsed_columns = self.get_time_series_parsed_columns()

    def get_database_columns(self):
        return [col["name"] for col in self.parsed_columns if "source" in col and col["source"] == "database"]

    # def get_method_columns(self):
    #     return [col['name'] for col in self.parsed_columns if col['type'] == 'method']

    def get_list_display_columns(self):
        columns = self.parsed_columns
        if self.time_series_pattern:
            time_series_columns = self.get_time_series_parsed_columns()
            try:
                index = self.columns.index("__time_series__")
                columns[index:index] = time_series_columns
            except ValueError:
                columns += time_series_columns

        if self.crosstab_field:
            crosstab_columns = self.get_crosstab_parsed_columns()

            try:
                index = self.columns.index("__crosstab__")
                columns[index:index] = crosstab_columns
            except ValueError:
                columns += crosstab_columns

        return columns

    def get_time_series_parsed_columns(self):
        """
        Return time series columns with all needed data attached
        :param plain: if True it returns '__total__' instead of '__total_TS011212'
        :return: List if columns
        """
        _values = []

        cols = self.time_series_columns or []
        series = self._get_time_series_dates(self.time_series_pattern)

        for index, dt in enumerate(series):
            for col in cols:
                magic_field_class = None

                if isinstance(col, str):
                    magic_field_class = field_registry.get_field_by_name(col)
                elif issubclass(col, ComputationField):
                    magic_field_class = col

                _values.append(
                    {
                        "name": magic_field_class.name + "TS" + dt[1].strftime("%Y%m%d"),
                        "original_name": magic_field_class.name,
                        "verbose_name": self.get_time_series_field_verbose_name(magic_field_class, dt, index, series),
                        "ref": magic_field_class,
                        "start_date": dt[0],
                        "end_date": dt[1],
                        "source": "magic_field" if magic_field_class else "",
                        "is_summable": magic_field_class.is_summable,
                    }
                )

            # append the crosstab fields, if they exist, on the time_series
            if self._crosstab_parsed_columns:
                for parsed_col in self._crosstab_parsed_columns:
                    parsed_col = parsed_col.copy()
                    parsed_col["name"] = parsed_col["name"] + "TS" + dt[1].strftime("%Y%m%d")
                    parsed_col["start_date"] = dt[0]
                    parsed_col["end_date"] = dt[1]
                    _values.append(parsed_col)

        return _values

    def get_time_series_field_verbose_name(self, computation_class, date_period, index, series, pattern=None):
        """
        Sent the column data to construct a verbose name.
        Default implementation is delegated to the ReportField.get_time_series_field_verbose_name
        (which is  name + the end date %Y%m%d)

        :param computation_class: the computation field_name
        :param date_period: a tuple of (start_date, end_date)
        :return: a verbose string
        """
        pattern = pattern or self.time_series_pattern
        return computation_class.get_time_series_field_verbose_name(date_period, index, series, pattern)

    def get_custom_time_series_dates(self):
        """
        Hook to get custom , maybe separated date periods
        :return: [ (date1,date2) , (date3,date4), .... ]
        """
        return self.time_series_custom_dates or []

    def _get_time_series_dates(self, series=None, start_date=None, end_date=None):
        from dateutil.relativedelta import relativedelta

        series = series or self.time_series_pattern
        start_date = start_date or self.start_date
        end_date = end_date or self.end_date
        _values = []

        if series:
            if series == "daily":
                time_delta = datetime.timedelta(days=1)
            elif series == "weekly":
                time_delta = relativedelta(weeks=1)
            elif series == "bi-weekly":
                time_delta = relativedelta(weeks=2)
            elif series == "monthly":
                time_delta = relativedelta(months=1)
            elif series == "quarterly":
                time_delta = relativedelta(months=3)
            elif series == "semiannually":
                time_delta = relativedelta(months=6)
            elif series == "annually":
                time_delta = relativedelta(years=1)
            elif series == "custom":
                return self.get_custom_time_series_dates()
            else:
                raise NotImplementedError(f'"{series}" is not implemented for time_series_pattern')

            done = False

            while not done:
                to_date = start_date + time_delta
                _values.append((start_date, to_date))
                start_date = to_date
                if to_date >= end_date:
                    done = True
        return _values

    def get_crosstab_parsed_columns(self):
        """
        Return a list of the columns analyzed , with reference to computation field and everything
        :return:
        """
        report_columns = self.crosstab_columns or []

        ids = list(self.crosstab_ids) or list(self.crosstab_ids_custom_filters)
        if self.crosstab_compute_remainder and not self.crosstab_ids_custom_filters:
            ids.append("----")
        output_cols = []

        ids_length = len(ids) - 1
        for counter, crosstab_id in enumerate(ids):
            queryset_filters = None

            if self.crosstab_ids_custom_filters:
                queryset_filters = crosstab_id
                crosstab_id = counter

            for col in report_columns:
                magic_field_class = None
                if isinstance(col, str):
                    magic_field_class = field_registry.get_field_by_name(col)
                elif issubclass(col, ComputationField):
                    magic_field_class = col

                crosstab_column = {
                    "name": f"{magic_field_class.name}CT{crosstab_id}",
                    "original_name": magic_field_class.name,
                    "verbose_name": self.get_crosstab_field_verbose_name(
                        magic_field_class, self.crosstab_field, crosstab_id
                    ),
                    "ref": magic_field_class,
                    "id": crosstab_id,
                    "crosstab_field": self.crosstab_field,
                    "is_remainder": counter == ids_length if self.crosstab_compute_remainder else False,
                    "source": "magic_field" if magic_field_class else "",
                    "is_summable": magic_field_class.is_summable,
                    "computation_flag": "crosstab",  # a flag, todo find a better way probably
                }
                crosstab_column["queryset_filters"] = self._construct_crosstab_filter(crosstab_column, queryset_filters)

                output_cols.append(crosstab_column)

        return output_cols

    def get_crosstab_field_verbose_name(self, computation_class, model, id):
        """
        Hook to change the crosstab field verbose name, default it delegate this function to the ReportField
        :param computation_class: ReportField Class
        :param model: the model name as string
        :param id: the current crosstab id
        :return: a verbose string
        """
        return computation_class.get_crosstab_field_verbose_name(model, id)

    def get_metadata(self):
        """
        A hook to send data about the report for front end which can later be used in charting
        :return:
        """
        time_series_columns = self.get_time_series_parsed_columns()
        crosstab_columns = self.get_crosstab_parsed_columns()
        metadata = {
            "time_series_pattern": self.time_series_pattern,
            "time_series_column_names": [x["name"] for x in time_series_columns],
            "time_series_column_verbose_names": [x["verbose_name"] for x in time_series_columns],
            "crosstab_model": self.crosstab_field or "",
            "crosstab_column_names": [x["name"] for x in crosstab_columns],
            "crosstab_column_verbose_names": [x["verbose_name"] for x in crosstab_columns],
        }
        return metadata

    def get_columns_data(self):
        """
        Hook to get the columns information to front end
        :param columns:
        :return:
        """
        columns = self.get_list_display_columns()
        data = []

        for col in columns:
            data.append(
                {
                    "name": col["name"],
                    "computation_field": col.get("original_name", ""),
                    "verbose_name": col["verbose_name"],
                    "visible": col.get("visible", True),
                    "type": col.get("type", "text"),
                    "is_summable": col.get("is_summable", ""),
                }
            )
        return data

    def get_full_response(
        self, data=None, report_slug=None, chart_settings=None, default_chart_title=None, default_chart_engine=None
    ):
        data = data or self.get_report_data()
        data = {
            "report_slug": report_slug or self.__class__.__name__,
            "data": data,
            "columns": self.get_columns_data(),
            "metadata": self.get_metadata(),
            "chart_settings": self.get_chart_settings(
                chart_settings, default_chart_title=default_chart_title, chart_engine=default_chart_engine
            ),
        }
        return data

    @staticmethod
    def get_chart_settings(chart_settings=None, default_chart_title=None, chart_engine=None):
        """
        Ensure the sane settings are passed to the front end.
        """
        chart_engine = chart_engine or SLICK_REPORTING_DEFAULT_CHARTS_ENGINE
        output = []
        chart_settings = chart_settings or []
        report_title = default_chart_title or ""
        for i, chart in enumerate(chart_settings):
            if type(chart) is Chart:
                chart = chart.to_dict()
            chart["id"] = chart.get("id", f"{i}")
            chart_type = chart.get("type", "line")
            if chart_type == "column" and SLICK_REPORTING_DEFAULT_CHARTS_ENGINE == "chartsjs":
                chart["type"] = "bar"

            if not chart.get("title", False):
                chart["title"] = report_title
            chart["engine_name"] = chart.get("engine_name", chart_engine)
            chart["entryPoint"] = (
                chart.get("entryPoint")
                or app_settings.SLICK_REPORTING_SETTINGS["CHARTS"][chart["engine_name"]]["entryPoint"]
            )

            output.append(chart)
        return output


class ListViewReportGenerator(ReportGenerator):
    def prepare_queryset(self, queryset):
        return self._apply_queryset_options(queryset, self.get_database_columns())

    def _apply_queryset_options(self, query, fields=None):
        """
        Apply the filters to the main queryset which will computed results be mapped to
        :param query:
        :param fields:
        :return:
        """
        filters = {}
        if self.date_field:
            filters = {
                f"{self.date_field}__gt": self.start_date,
                f"{self.date_field}__lte": self.end_date,
            }
        filters.update(self.kwargs_filters)

        if filters:
            query = query.filter(**filters)
        if fields:
            return query.values(*fields)
        return query

    def _get_record_data(self, obj, columns):
        """
        the function is run for every obj in the main_queryset
        :param obj: current row
        :param: columns： The columns we iterate on
        :return: a dict object containing all needed data
        """

        data = {}
        group_by_val = None
        if self.group_by:
            if self.group_by_field.related_model and "__" not in self.group_by:
                primary_key_name = self.get_primary_key_name(self.group_by_field.related_model)
            else:
                primary_key_name = self.group_by_field_attname

            column_data = obj.get(primary_key_name, obj.get("id"))
            group_by_val = str(column_data)

        for window, window_cols in columns:
            for col_data in window_cols:
                name = col_data["name"]

                if col_data.get("source", "") == "attribute_field":
                    data[name] = col_data["ref"](self, obj, data)
                    # changed line
                elif col_data.get("source", "") == "container_class_attribute_field":
                    data[name] = col_data["ref"](obj)

                elif (col_data.get("source", "") == "magic_field" and self.group_by) or (
                    self.time_series_pattern and not self.group_by
                ):
                    source = self._report_fields_dependencies[window].get(name, False)
                    if source:
                        computation_class = self.report_fields_classes[source]
                        value = computation_class.get_dependency_value(group_by_val, col_data["ref"].name)
                    else:
                        try:
                            computation_class = self.report_fields_classes[name]
                        except KeyError:
                            continue
                        value = computation_class.do_resolve(group_by_val, data)
                    if self.swap_sign:
                        value = -value
                    data[name] = value

                else:
                    data[name] = obj[name]
        return data

    def _remove_order(self, main_queryset):
        return main_queryset



def report_field_register(report_field, *args, **kwargs):
    """
    Registers the given model(s) classes and wrapped ModelAdmin class with
    admin site:

    @register(Author)
    class AuthorAdmin(admin.ModelAdmin):
        pass

    A kwarg of `site` can be passed as the admin site, otherwise the default
    admin site will be used.
    """
    from .fields import ComputationField
    from .registry import field_registry

    def _model_admin_wrapper(admin_class):
        if not issubclass(admin_class, ComputationField):
            raise ValueError("Wrapped class must subclass ComputationField.")

        field_registry.register(report_field)

    _model_admin_wrapper(report_field)
    return report_field



import csv
import datetime
import warnings

import simplejson as json
from django import forms
from django.conf import settings
from django.db.models import Q
from django.forms import modelform_factory
from django.http import HttpResponse, StreamingHttpResponse, JsonResponse
from django.utils.encoding import force_str
from django.utils.functional import Promise
from django.views.generic import FormView

from .app_settings import SLICK_REPORTING_SETTINGS
from .forms import (
    report_form_factory,
    get_crispy_helper,
    default_formfield_callback,
    OrderByForm,
)
from .generator import (
    ReportGenerator,
    ListViewReportGenerator,
    ReportGeneratorAPI,
    Chart,  # noqa # needed for easier importing in other apps
)


def dictsort(value, arg, desc=False):
    """
    Takes a list of dicts, returns that list sorted by the property given in
    the argument.
    """
    return sorted(value, key=lambda x: x[arg], reverse=desc)


class ExportToCSV(object):
    def get_filename(self):
        return self.report_title

    def get_response(self):
        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = "attachment; filename={filename}.csv".format(filename=self.get_filename())

        writer = csv.writer(response)
        for rows in self.get_rows():
            writer.writerow(rows)

        return response

    def get_rows(self):
        columns, verbose_names = self.get_columns()
        yield verbose_names
        for line in self.report_data["data"]:
            yield [line[col_name] for col_name in columns]

    def get_columns(self, extra_context=None):
        return list(zip(*[(x["name"], x["verbose_name"]) for x in self.report_data["columns"]]))

    def __init__(self, request, report_data, report_title, **kwargs):
        self.request = request
        self.report_data = report_data
        self.report_title = report_title
        self.kwargs = kwargs


class ExportToStreamingCSV(ExportToCSV):
    def get_response(self):
        # Copied form Djagno Docs
        class Echo:
            def write(self, value):
                return value

        pseudo_buffer = Echo()
        writer = csv.writer(pseudo_buffer)
        return StreamingHttpResponse(
            (writer.writerow(row) for row in self.get_rows()),
            content_type="text/csv",
            headers={
                "Content-Disposition": 'attachment; filename="{filename}.csv"'.format(filename=self.get_filename())
            },
        )


class ReportViewBase(ReportGeneratorAPI, FormView):
    report_slug = None

    report_title = ""

    report_title_context_key = "report_title"

    report_generator_class = ReportGenerator

    base_model = None

    chart_settings = None

    excluded_fields = None

    time_series_selector = False
    time_series_selector_choices = None
    time_series_selector_default = None
    time_series_selector_allow_empty = False

    csv_export_class = ExportToStreamingCSV

    with_type = False
    doc_type_field_name = "doc_type"
    doc_type_plus_list = None
    doc_type_minus_list = None
    auto_load = True
    chart_engine = ""

    default_order_by = ""

    template_name = "slick_reporting/report.html"

    export_actions = None

    @staticmethod
    def form_filter_func(fkeys_dict):
        # todo revise
        return fkeys_dict

    @classmethod
    def get_report_title(cls):
        """
        :return: The report name
        """
        name = cls.__name__
        if cls.report_title:
            name = cls.report_title
        return name

    def order_results(self, data):
        """
        order the results based on GET parameter or default_order_by
        :param data: List of Dict to be ordered
        :return: Ordered data
        """
        order_field, asc = OrderByForm(self.request.GET).get_order_by(self.default_order_by)
        if order_field:
            data = dictsort(data, order_field, asc)
        return data

    def get_doc_types_q_filters(self):
        if self.doc_type_plus_list or self.doc_type_minus_list:
            return (
                [Q(**{f"{self.doc_type_field_name}__in": self.doc_type_plus_list})] if self.doc_type_plus_list else []
            ), (
                [Q(**{f"{self.doc_type_field_name}__in": self.doc_type_minus_list})] if self.doc_type_minus_list else []
            )

        return [], []

    def get_export_actions(self):
        """
        Hook to get the export options
        :return: list of export options
        """
        actions = ["export_csv"] if self.csv_export_class else []

        if self.export_actions:
            actions = actions + self.export_actions

        export_actions = []

        for action in actions:
            func = getattr(self, action, None)
            parameter = action.replace("export_", "")

            export_actions.append(
                {
                    "name": action,
                    "title": getattr(func, "title", action.replace("_", " ").title()),
                    "icon": getattr(func, "icon", ""),
                    "css_class": getattr(func, "css_class", ""),
                    "parameter": parameter,
                }
            )
        return export_actions

    def get(self, request, *args, **kwargs):
        form_class = self.get_form_class()
        self.form = self.get_form(form_class)
        report_data = {}
        if self.form.is_valid():
            if self.request.GET or self.request.POST or request.headers.get("x-requested-with") == "XMLHttpRequest":
                # only display results if it's requested,
                # considered requested if it's ajax request, or a populated GET or POST.
                report_data = self.get_report_results()

                export_option = request.GET.get("_export", "")
                if export_option:
                    try:
                        return getattr(self, f"export_{export_option}")(report_data)
                    except AttributeError:
                        pass

                if request.headers.get("x-requested-with") == "XMLHttpRequest":
                    return self.ajax_render_to_response(report_data)

            return self.render_to_response(self.get_context_data(report_data=report_data))
        else:
            return self.form_invalid(self.form)

        # return self.render_to_response(self.get_context_data())

    def export_csv(self, report_data):
        return self.csv_export_class(self.request, report_data, self.report_title).get_response()

    export_csv.title = SLICK_REPORTING_SETTINGS["MESSAGES"]["export_to_csv"]
    export_csv.css_class = "btn btn-primary"
    export_csv.icon = ""

    @classmethod
    def get_report_model(cls):
        if cls.queryset is not None:
            return cls.queryset.model
        return cls.report_model

    def ajax_render_to_response(self, report_data):
        return HttpResponse(self.serialize_to_json(report_data), content_type="application/json")

    def serialize_to_json(self, response_data):
        """Returns the JSON string for the compiled data object."""

        def date_handler(obj):
            if type(obj) is datetime.datetime:
                return obj.strftime("%Y-%m-%d %H:%M")
            elif hasattr(obj, "isoformat"):
                return obj.isoformat()
            elif isinstance(obj, Promise):
                return force_str(obj)

        indent = None
        if settings.DEBUG:
            indent = 4

        return json.dumps(response_data, indent=indent, use_decimal=True, default=date_handler)

    def get_form_class(self):
        """
        Automatically instantiate a form based on details provided
        :return:
        """
        return self.form_class or report_form_factory(
            self.get_report_model(),
            crosstab_model=self.crosstab_field,
            display_compute_remainder=self.crosstab_compute_remainder,
            excluded_fields=self.excluded_fields,
            fkeys_filter_func=self.form_filter_func,
            initial=self.get_initial(),
            show_time_series_selector=self.time_series_selector,
            time_series_selector_choices=self.time_series_selector_choices,
            time_series_selector_default=self.time_series_selector_default,
            time_series_selector_allow_empty=self.time_series_selector_allow_empty,
        )

    def get_form_kwargs(self):
        """
        Returns the keyword arguments for instantiating the form.
        """
        kwargs = {
            "initial": self.get_initial(),
            "prefix": self.get_prefix(),
        }

        if self.request.method in ("POST", "PUT"):
            kwargs.update(
                {
                    "data": self.request.POST,
                    "files": self.request.FILES,
                }
            )
        elif self.request.method in ("GET", "PUT"):
            if self.request.GET or self.request.headers.get("x-requested-with") == "XMLHttpRequest":
                kwargs.update(
                    {
                        "data": self.request.GET,
                    }
                )
        return kwargs

    def get_crosstab_ids(self):
        """
        Hook to get the crosstab ids
        :return:
        """
        return self.form.get_crosstab_ids()

    def get_group_by_custom_querysets(self):
        return self.group_by_custom_querysets

    def get_report_generator(self, queryset=None, for_print=False):
        queryset = queryset or self.get_queryset()
        q_filters, kw_filters = self.form.get_filters()
        crosstab_compute_remainder = False
        if self.crosstab_field:
            self.crosstab_ids = self.get_crosstab_ids()
        try:
            crosstab_compute_remainder = (
                self.form.get_crosstab_compute_remainder()
                if self.request.GET or self.request.POST
                else self.crosstab_compute_remainder
            )
        except NotImplementedError:
            pass

        time_series_pattern = self.time_series_pattern
        if self.time_series_selector:
            time_series_pattern = self.form.get_time_series_pattern()

        doc_type_plus_list, doc_type_minus_list = [], []

        if self.with_type:
            doc_type_plus_list, doc_type_minus_list = self.get_doc_types_q_filters()

        return self.report_generator_class(
            self.get_report_model(),
            start_date=self.form.get_start_date(),
            end_date=self.form.get_end_date(),
            q_filters=q_filters,
            kwargs_filters=kw_filters,
            date_field=self.date_field,
            main_queryset=queryset,
            print_flag=for_print,
            limit_records=self.limit_records,
            swap_sign=self.swap_sign,
            columns=self.columns,
            group_by=self.group_by,
            group_by_custom_querysets=self.get_group_by_custom_querysets(),
            group_by_custom_querysets_column_verbose_name=self.group_by_custom_querysets_column_verbose_name,
            time_series_pattern=time_series_pattern,
            time_series_columns=self.time_series_columns,
            time_series_custom_dates=self.time_series_custom_dates,
            crosstab_field=self.crosstab_field,
            crosstab_ids=self.crosstab_ids,
            crosstab_columns=self.crosstab_columns,
            crosstab_compute_remainder=crosstab_compute_remainder,
            crosstab_ids_custom_filters=self.crosstab_ids_custom_filters,
            format_row_func=self.format_row,
            container_class=self,
            doc_type_plus_list=doc_type_plus_list,
            doc_type_minus_list=doc_type_minus_list,
        )

    def format_row(self, row_obj):
        """
        A hook to format each row . This method gets called on each row in the results. <ust return the object
        :param row_obj: a dict representing a single row in the results
        :return: A dict representing a single row in the results
        """
        return row_obj

    @classmethod
    def get_columns_data(cls, generator):
        """
        Hook to get the columns information to front end
        :param generator: the SlickReportGenerator instance used
        :return:
        """
        return generator.get_columns_data()

    def get_report_results(self, for_print=False):
        """
        Gets the reports Data, and, its meta data used by datatables.net and highcharts
        :return: JsonResponse
        """

        queryset = self.get_queryset()
        report_generator = self.get_report_generator(queryset, for_print)
        data = report_generator.get_report_data()
        data = self.filter_results(data, for_print)
        data = self.order_results(data)

        return report_generator.get_full_response(
            data=data,
            report_slug=self.get_report_slug(),
            chart_settings=self.chart_settings,
            default_chart_title=self.report_title,
            default_chart_engine=self.chart_engine,
        )

    @classmethod
    def get_metadata(cls, generator):
        """
        A hook to send data about the report for front end which can later be used in charting
        :return:
        """
        return generator.get_metadata()

    def get_chart_settings(self, generator=None):
        """
        Ensure the sane settings are passed to the front end.
        """
        return self.report_generator_class.get_chart_settings(
            chart_settings=self.chart_settings or [],
            default_chart_title=self.report_title,
            chart_engine=self.chart_engine,
        )

    @classmethod
    def get_queryset(cls):
        if cls.queryset is None:
            return cls.get_report_model()._default_manager.all()
        return cls.queryset

    def filter_results(self, data, for_print=False):
        """
        Hook to Filter results based on computed data (like eliminate __balance__ = 0, etc)
        return None to remove the row from the results
        :param data: List of objects
        :param for_print: is print request
        :return: filtered data
        """
        return data

    @classmethod
    def get_report_slug(cls):
        return cls.report_slug or cls.__name__.lower()

    def get_initial(self):
        initial = self.initial.copy()
        initial.update(
            {
                "start_date": SLICK_REPORTING_SETTINGS["DEFAULT_START_DATE_TIME"],
                "end_date": SLICK_REPORTING_SETTINGS["DEFAULT_END_DATE_TIME"],
            }
        )
        return initial

    def get_form_crispy_helper(self):
        """
        A hook retuning crispy helper for the form
        :return:
        """
        if hasattr(self, "form"):
            return self.form.get_crispy_helper()
        return None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context[self.report_title_context_key] = self.report_title
        context["crispy_helper"] = self.get_form_crispy_helper()
        context["auto_load"] = self.auto_load
        context["report"] = self

        if not (self.request.POST or self.request.GET):
            context["form"] = self.get_form_class()(**self.get_form_kwargs())
        return context

    def form_invalid(self, form):
        if self.request.META.get("HTTP_X_REQUESTED_WITH") == "XMLHttpRequest":
            return JsonResponse(form.errors, status=400)
        return super().form_invalid(form)


class ReportView(ReportViewBase):
    def __init_subclass__(cls) -> None:
        # date_field = getattr(cls, 'date_field', '')
        # if not date_field:
        #     raise TypeError(f'`date_field` is not set on {cls}')
        # cls.report_generator_class.check_columns([cls.date_field], False, cls.get_report_model())

        # sanity check, raises error if the columns or date fields is not set
        if cls.columns:
            cls.report_generator_class.check_columns(
                cls,
                cls.columns,
                cls.group_by,
                cls.get_report_model(),
                container_class=cls,
                group_by_custom_querysets=cls.group_by_custom_querysets,
            )

        super().__init_subclass__()


class SlickReportingListViewMixin(ReportViewBase):
    report_generator_class = ListViewReportGenerator
    filters = None

    def get_queryset(self):
        qs = self.queryset or self.report_model.objects
        if self.default_order_by:
            qs.order_by(self.default_order_by)
        return qs

    def get_form_filters(self, form):
        if self.form_class:
            return form.get_filters()

        kw_filters = {}

        for name, field in form.base_fields.items():
            if type(field) is forms.ModelMultipleChoiceField:
                value = form.cleaned_data[name]
                if value:
                    kw_filters[f"{name}__in"] = form.cleaned_data[name]
            elif type(field) is forms.BooleanField:
                # boolean field while checked on frontend , and have initial = True, give false value on cleaned_data
                #  Hence this check to see if it was indeed in the GET params,
                value = field.initial
                if self.request.GET:
                    value = form.cleaned_data.get(name, False)
                kw_filters[name] = value

            else:
                value = form.cleaned_data[name]
                if value:
                    kw_filters[name] = form.cleaned_data[name]

        return [], kw_filters

    def get_form_crispy_helper(self):
        return get_crispy_helper(self.filters)

    def get_report_generator(self, queryset=None, for_print=False):
        q_filters, kw_filters = self.get_form_filters(self.form)

        return self.report_generator_class(
            self.get_report_model(),
            # start_date=self.form.get_start_date(),
            # end_date=self.form.get_end_date(),
            q_filters=q_filters,
            kwargs_filters=kw_filters,
            date_field=self.date_field,
            main_queryset=queryset,
            print_flag=for_print,
            limit_records=self.limit_records,
            columns=self.columns,
            format_row_func=self.format_row,
            container_class=self,
        )

    def get_form_class(self):
        if self.form_class:
            return self.form_class

        elif self.filters:
            return modelform_factory(
                model=self.get_report_model(),
                fields=self.filters,
                formfield_callback=default_formfield_callback,
            )
        return forms.Form

    def get_report_results(self, for_print=False):
        """
        Gets the reports Data, and, its meta data used by datatables.net and highcharts
        :return: JsonResponse
        """

        queryset = self.get_queryset()
        report_generator = self.get_report_generator(queryset, for_print)
        data = report_generator.get_report_data()
        data = self.filter_results(data, for_print)

        return report_generator.get_full_response(
            data=data,
            report_slug=self.get_report_slug(),
            chart_settings=self.chart_settings,
            default_chart_title=self.report_title,
        )


class SlickReportingListView(SlickReportingListViewMixin, ReportViewBase):
    def __init_subclass__(cls) -> None:
        warnings.warn(
            "slick_reporting.view.SlickReportingListView is"
            "deprecated in favor of slick_reporting.view.ListReportView",
            Warning,
            stacklevel=2,
        )
        super().__init_subclass__()


class ListReportView(SlickReportingListViewMixin):
    pass


class SlickReportViewBase(ReportViewBase):
    """
    Deprecated in favor of slick_reporting.view.ReportViewBase
    """

    def __init_subclass__(cls) -> None:
        warnings.warn(
            "slick_reporting.view.SlickReportView and slick_reporting.view.SlickReportViewBase are "
            "deprecated in favor of slick_reporting.view.ReportView and slick_reporting.view.BaseReportView",
            Warning,
            stacklevel=2,
        )

        super().__init_subclass__()


class SlickReportView(ReportView):
    def __init_subclass__(cls) -> None:
        warnings.warn(
            "slick_reporting.view.SlickReportView and slick_reporting.view.SlickReportViewBase are "
            "deprecated in favor of slick_reporting.view.ReportView and slick_reporting.view.BaseReportView",
            Warning,
            stacklevel=2,
        )

        # cls.report_generator_class.check_columns(
        #     cls.columns, cls.group_by, cls.get_report_model(), container_class=cls
        # )



from __future__ import annotations

from warnings import warn

from django.db.models import Sum, Q
from django.template.defaultfilters import date as date_filter
from django.utils.translation import gettext_lazy as _

from .registry import field_registry


class ComputationField(object):
    """
    Computation field responsible for making the calculation unit
    """

    _field_registry = field_registry
    name = ""
    """The name to be used in the ReportGenerator"""

    calculation_field = "value"
    """the Field to compute on"""

    calculation_method = Sum
    """The computation Method"""

    verbose_name = None
    """Verbose name to be used in front end when needed"""

    requires = None
    """This can be a list of sibling classes,
    they will be asked to compute and their value would be available to you in the `resolve` method
    requires = [BasicCalculationA, BasicCalculationB]
    """

    type = "number"
    """Just a string describing what this computation field return, usually passed to frontend"""

    is_summable = True
    """Indicate if this computation can be summed over. Useful to be passed to frontend or whenever needed"""

    report_model = None
    """The model on which the computation would occur"""

    queryset = None
    """The queryset on which the computation would occur"""

    group_by = None
    group_by_custom_querysets = None
    plus_side_q = None
    minus_side_q = None

    base_kwargs_filters = None
    base_q_filters = None

    _require_classes = None
    _debit_and_credit = True

    prevent_group_by = False
    """Will prevent group by calculation for this specific field, serves when you want to compute overall results"""

    def __new__(cls, *args, **kwargs):
        """
        This is where we register the class in the registry
        :param args:
        :param kwargs:
        :return:
        """
        if not cls.name:
            raise ValueError(f"ReportField {cls} must have a name")
        return super(ComputationField, cls).__new__(cls)

    @classmethod
    def create(cls, method, field, name=None, verbose_name=None, is_summable=True):
        """
        Creates a ReportField class on the fly
        :param method: The computation Method to be used
        :param field: The field on which the computation would occur
        :param name: a name to refer to this field else where
        :param verbose_name: Verbose name
        :param is_summable:
        :return:
        """
        if not name:
            name = name or f"{method.name.lower()}__{field}"
            assert name not in cls._field_registry.get_all_report_fields_names()

        verbose_name = verbose_name or f"{method.name} {field}"
        report_klass = type(
            f"ReportField_{name}",
            (cls,),
            {
                "name": name,
                "verbose_name": verbose_name,
                "calculation_field": field,
                "calculation_method": method,
                "is_summable": is_summable,
            },
        )
        return report_klass

    def __init__(
        self,
        plus_side_q=None,
        minus_side_q=None,
        report_model=None,
        queryset=None,
        calculation_field=None,
        calculation_method=None,
        date_field="",
        group_by=None,
        group_by_custom_querysets=None,
    ):
        super(ComputationField, self).__init__()
        self.date_field = date_field
        self.report_model = self.report_model or report_model
        self.queryset = self.queryset or queryset
        self.queryset = self.report_model._default_manager.all() if self.queryset is None else self.queryset

        self.group_by_custom_querysets = self.group_by_custom_querysets or group_by_custom_querysets

        self.calculation_field = calculation_field if calculation_field else self.calculation_field
        self.calculation_method = calculation_method if calculation_method else self.calculation_method
        self.plus_side_q = self.plus_side_q or plus_side_q
        self.minus_side_q = self.minus_side_q or minus_side_q
        self.requires = self.requires or []
        self.group_by = self.group_by or group_by
        self._cache = None, None, None
        self._require_classes = self._get_required_classes()
        self._required_prepared_results = None

        self._debit_and_credit = self.plus_side_q or self.minus_side_q

    @classmethod
    def _get_required_classes(cls):
        requires = cls.requires or []
        return [field_registry.get_field_by_name(x) if isinstance(x, str) else x for x in requires]

    def apply_aggregation(self, queryset, group_by=""):
        annotation = self.calculation_method(self.calculation_field)
        if self.group_by_custom_querysets:
            return queryset.aggregate(annotation)
        elif group_by:
            queryset = queryset.values(group_by).annotate(annotation)
            queryset = {str(x[self.group_by]): x for x in queryset}
        else:
            queryset = queryset.aggregate(annotation)
        return queryset

    def init_preparation(self, q_filters=None, kwargs_filters=None, **kwargs):
        """
        Called by the generator to prepare the calculation of this field + it's requirements
        :param q_filters:
        :param kwargs_filters:
        :param kwargs:
        :return:
        """
        kwargs_filters = kwargs_filters or {}

        required_prepared_results = self._prepare_required_computations(q_filters, kwargs_filters.copy())
        queryset = self.get_queryset()
        if self.group_by_custom_querysets:
            debit_results, credit_results = self.prepare_custom_group_by_queryset(q_filters, kwargs_filters, **kwargs)
        else:
            debit_results, credit_results = self.prepare(
                q_filters,
                kwargs_filters,
                queryset,
                self.group_by,
                self.prevent_group_by,
                **kwargs,
            )
        self._cache = debit_results, credit_results
        self._required_prepared_results = required_prepared_results

    def prepare_custom_group_by_queryset(self, q_filters=None, kwargs_filters=None, **kwargs):
        debit_output, credit_output = [], []
        for index, queryset in enumerate(self.group_by_custom_querysets):
            debit, credit = self.prepare(q_filters, kwargs_filters, queryset, **kwargs)
            if debit:
                debit_output.append(debit)
            if credit:
                credit_output.append(credit)
        return debit_output, credit_output

    def prepare(
        self,
        q_filters: list | object = None,
        kwargs_filters: dict = None,
        main_queryset=None,
        group_by: str = None,
        prevent_group_by=None,
        **kwargs,
    ):
        """
        This is the first hook where you can customize the calculation away from the Django Query aggregation method
        This method is called with all available arguments, so you can prepare the results for the whole set and save
        it in a local cache (like self._cache) .
        The flow will later call the method `resolve`,  giving you the id, for you to return it respective calculation

        :param q_filters:
        :param kwargs_filters:
        :param main_queryset:
        :param group_by:
        :param prevent_group_by:
        :param kwargs:
        :return:
        """

        queryset = main_queryset.all()
        group_by = "" if prevent_group_by else group_by
        credit_results = None

        if q_filters:
            if type(q_filters) is Q:
                q_filters = [q_filters]
            queryset = queryset.filter(*q_filters)
        if kwargs_filters:
            queryset = queryset.filter(**kwargs_filters)

        if self.plus_side_q:
            queryset = queryset.filter(*self.plus_side_q)
        debit_results = self.apply_aggregation(queryset, group_by)

        if self._debit_and_credit:
            queryset = main_queryset.all()
            if kwargs_filters:
                queryset = queryset.filter(**kwargs_filters)
            if q_filters:
                queryset = queryset.filter(*q_filters)
            if self.minus_side_q:
                queryset = queryset.filter(*self.minus_side_q)
            credit_results = self.apply_aggregation(queryset, group_by)

        return debit_results, credit_results

    def get_queryset(self):
        queryset = self.queryset
        if self.base_q_filters:
            queryset = queryset.filter(*self.base_q_filters)
        if self.base_kwargs_filters:
            queryset = queryset.filter(**self.base_kwargs_filters)
        return queryset.order_by()

    def _prepare_required_computations(
        self,
        q_filters=None,
        extra_filters=None,
    ):
        values = {}
        for required_klass in self._require_classes:
            dep = required_klass(
                self.plus_side_q,
                self.minus_side_q,
                self.report_model,
                date_field=self.date_field,
                group_by=self.group_by,
                queryset=self.queryset,
                group_by_custom_querysets=self.group_by_custom_querysets,
            )
            results = dep.init_preparation(q_filters, extra_filters)
            values[dep.name] = {"results": results, "instance": dep}
        return values

    def resolve(self, prepared_results, required_computation_results: dict, current_pk, current_row=None) -> float:
        """
        Reponsible for getting the exact data from the prepared value
        :param prepared_results: the returned data from prepare
        :param required_computation_results: the returned data from prepare
        :param current_pk: he value of group by id
        :param current_row: the row in iteration
        :return: a solid number or value
        """
        debit_value, credit_value = self.extract_data(prepared_results, current_pk)
        value = debit_value or 0 - credit_value or 0
        return value

    def do_resolve(self, current_obj, current_row=None):
        prepared_result = self._cache
        dependencies_value = self._resolve_dependencies(current_obj)
        return self.resolve(prepared_result, dependencies_value, current_obj, current_row)

    def get_dependency_value(self, current_obj, name):
        """
        Get the values of the ReportFields specified in `requires`

        :param current_obj: the current object which we want the calculation for
        :param name: the name of the specific dependency you want.

        :return: a dict containing dependencies names as keys and their calculation as values
                 or a specific value if name is specified.
        """
        values = self._resolve_dependencies(current_obj, name=name)
        return values.get(name)

    def _resolve_dependencies(self, current_obj, name=None):
        dep_results = {}
        dependencies_value = self._required_prepared_results
        dependencies_value = dependencies_value or {}
        needed_values = [name] if name else dependencies_value.keys()
        for d in needed_values:
            d_instance = dependencies_value[d]["instance"]
            dep_results[d] = d_instance.do_resolve(current_obj)
        return dep_results

    def extract_data(self, prepared_results, current_obj):
        group_by = "" if self.prevent_group_by else (self.group_by or self.group_by_custom_querysets)
        annotation = "__".join([self.calculation_field.lower(), self.calculation_method.name.lower()])

        cached_debit, cached_credit = prepared_results

        cached = [cached_debit, cached_credit]
        output = []
        for results in cached:
            value = 0
            if results:
                if not group_by:
                    x = list(results.keys())[0]
                    value = results[x]
                elif self.group_by_custom_querysets:
                    value = results[int(current_obj)][annotation]
                else:
                    value = results.get(str(current_obj), {}).get(annotation, 0)
            output.append(value)
        return output

    @classmethod
    def get_full_dependency_list(cls):
        """
        Get the full Hirearchy of dependencies and dependencies dependency.
        :return: List of dependecies classes
        """

        def get_dependency(field_class):
            dependencies = field_class._get_required_classes()
            klasses = []
            for klass in dependencies:
                klasses.append(klass)
                other = get_dependency(klass)
                if other:
                    klasses += other
            return klasses

        return get_dependency(cls)

    @classmethod
    def get_crosstab_field_verbose_name(cls, model, id):
        """
        Construct a verbose name for the crosstab field
        :param model: the model name
        :param id: the id of the current crosstab object
        :return: a verbose string
        """
        if id == "----":
            return _("The remainder")
        return f"{cls.verbose_name} {model} {id}"

    @classmethod
    def get_time_series_field_verbose_name(cls, date_period, index, dates, pattern):
        """
        Get the name of the verbose name of a computation field that's in a time_series.
        should be a mix of the date period of the column and it's verbose name.
        :param date_period: a tuple of (start_date, end_date)
        :param index: the index of the current field in the whole dates to be calculated
        :param dates a list of tuples representing the start and the end date
        :param pattern it's the pattern name. monthly, daily, custom, ...
        :return: a verbose string
        """
        dt_format = "%Y/%m/%d"

        if pattern == "monthly":
            month_name = date_filter(date_period[0], "F Y")
            return f"{cls.verbose_name} {month_name}"
        elif pattern == "daily":
            return f"{cls.verbose_name} {date_period[0].strftime(dt_format)}"
        elif pattern == "weekly":
            return f' {cls.verbose_name} {_("Week")} {index + 1} {date_period[0].strftime(dt_format)}'
        elif pattern == "yearly":
            year = date_filter(date_period[0], "Y")
            return f"{cls.verbose_name} {year}"

        return f"{cls.verbose_name} {date_period[0].strftime(dt_format)} - {date_period[1].strftime(dt_format)}"


class FirstBalanceField(ComputationField):
    name = "__fb__"
    verbose_name = _("opening balance")

    def prepare(
        self,
        q_filters: list | object = None,
        kwargs_filters: dict = None,
        main_queryset=None,
        group_by: str = None,
        prevent_group_by=None,
        **kwargs,
    ):
        extra_filters = kwargs_filters or {}

        from_date_value = extra_filters.get(f"{self.date_field}__gte")
        extra_filters.pop(f"{self.date_field}__gte", None)
        extra_filters[f"{self.date_field}__lt"] = from_date_value
        return super(FirstBalanceField, self).prepare(
            q_filters, kwargs_filters, main_queryset, group_by, prevent_group_by, **kwargs
        )


field_registry.register(FirstBalanceField)


class TotalReportField(ComputationField):
    name = "__total__"
    verbose_name = _("Sum of value")
    requires = ["__debit__", "__credit__"]


field_registry.register(TotalReportField)


class BalanceReportField(ComputationField):
    name = "__balance__"
    verbose_name = _("Closing Total")
    requires = ["__fb__"]

    def resolve(self, prepared_results, required_computation_results: dict, current_pk, current_row=None) -> float:
        result = super().resolve(prepared_results, required_computation_results, current_pk, current_row)
        fb = required_computation_results.get("__fb__") or 0

        return result + fb


field_registry.register(BalanceReportField)


class PercentageToTotalBalance(ComputationField):
    requires = [BalanceReportField]
    name = "__percent_to_total_balance__"
    verbose_name = _("%")

    prevent_group_by = True

    def resolve(self, prepared_results, required_computation_results: dict, current_pk, current_row=None) -> float:
        result = super().resolve(prepared_results, required_computation_results, current_pk, current_row)
        return required_computation_results.get("__balance__") / result * 100


class CreditReportField(ComputationField):
    name = "__credit__"
    verbose_name = _("Credit")

    def resolve(self, prepared_results, required_computation_results: dict, current_pk, current_row=None) -> float:
        debit_value, credit_value = self.extract_data(prepared_results, current_pk)
        return credit_value


field_registry.register(CreditReportField)


@field_registry.register
class DebitReportField(ComputationField):
    name = "__debit__"
    verbose_name = _("Debit")

    def resolve(self, prepared_results, required_computation_results: dict, current_pk, current_row=None) -> float:
        debit_value, credit_value = self.extract_data(prepared_results, current_pk)
        return debit_value


@field_registry.register
class CreditQuantityReportField(ComputationField):
    name = "__credit_quantity__"
    verbose_name = _("Credit QTY")
    calculation_field = "quantity"
    is_summable = False

    def resolve(self, prepared_results, required_computation_results: dict, current_pk, current_row=None) -> float:
        debit_value, credit_value = self.extract_data(prepared_results, current_pk)
        return credit_value


@field_registry.register
class DebitQuantityReportField(ComputationField):
    name = "__debit_quantity__"
    calculation_field = "quantity"
    verbose_name = _("Debit QTY")
    is_summable = False

    def resolve(self, prepared_results, required_computation_results: dict, current_pk, current_row=None) -> float:
        debit_value, credit_value = self.extract_data(prepared_results, current_pk)
        return debit_value


class TotalQTYReportField(ComputationField):
    name = "__total_quantity__"
    verbose_name = _("Total QTY")
    calculation_field = "quantity"
    is_summable = False


field_registry.register(TotalQTYReportField)


class FirstBalanceQTYReportField(FirstBalanceField):
    name = "__fb_quantity__"
    verbose_name = _("Opening QTY")
    calculation_field = "quantity"
    is_summable = False


field_registry.register(FirstBalanceQTYReportField)


class BalanceQTYReportField(ComputationField):
    name = "__balance_quantity__"
    verbose_name = _("Closing QTY")
    calculation_field = "quantity"
    requires = ["__fb_quantity__"]
    is_summable = False

    def resolve(self, prepared_results, required_computation_results: dict, current_pk, current_row=None) -> float:
        result = super().resolve(prepared_results, required_computation_results, current_pk, current_row)
        fb = required_computation_results.get("__fb_quantity__") or 0
        return result + fb


field_registry.register(BalanceQTYReportField)


class SlickReportField(ComputationField):
    @staticmethod
    def warn():
        warn(
            "SlickReportField name is deprecated, please use ComputationField instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    @classmethod
    def create(cls, method, field, name=None, verbose_name=None, is_summable=True):
        cls.warn()
        return super().create(method, field, name, verbose_name, is_summable)

    def __new__(cls, *args, **kwargs):
        cls.warn()
        return super().__new__(cls, *args, **kwargs)



XXXXXXXXXX
XXXXXXXXXXX
XXXXXXXXXX
XXXXXXXXX
XXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX